from scopeton import compat, constants


def flatten(lst):
    res = []
    for el in lst:
        if type(el) is list:
            res += flatten(el)
        else:
            res.append(el)
    return res


def get_class_name(cls):
    return cls.__name__ if hasattr(cls, "__name__") else str(cls)


def get_class_tree_qualifiers(cls):
    res = [get_bean_qualifier(k) for k in get_class_tree(cls)]
    res.reverse()
    return res


def rm_dups(lst):
    res = []
    for k in lst:
        if k not in res:
            res.append(k)
    return res


def get_class_tree(cls):
    return rm_dups(_get_class_tree(cls))


def _get_class_tree(cls):
    res = [cls]
    if type(cls) is not object and hasattr(cls, "__bases__") and cls.__bases__:
        res += flatten([_get_class_tree(parent) for parent in cls.__bases__])
    return res


def get_bean_qualifier(bean) -> str:
    if isinstance(bean, str):
        return bean
    from scopeton.objects import Bean
    if isinstance(bean, Bean):
        return bean.qualifier_tree[-1]
    if not hasattr(bean, "__name__"):
        return str(bean)
    return bean.__name__


def call_method_by_name(obj, name, *args, **kwargs):
    if hasattr(obj, name):
        return getattr(obj, name)(*args, **kwargs)


def get_injectables(instance):
    return [fn for fn in compat.get_methods(instance) if hasattr(fn, constants.TO_INJECT_FLAG)]


def get_scope_by_method_or_class(target):
    if compat.is_method(target):
        target = compat.get_method_instance(target)
    if not hasattr(target, constants.SCOPE_PARAMETER):
        raise ScopetonException("No scope parameter: {} for instance {}".format(constants.SCOPE_PARAMETER, target))
    return getattr(target, constants.SCOPE_PARAMETER)


class ScopetonException(Exception):
    def __init__(self, *args, **kwargs):
        super(Exception, self).__init__(*args, **kwargs)


def set_instance_attr(instance, attr, value):
    if is_primitive(instance):
        return
    setattr(instance, attr, value)


def is_primitive(obj):
    return not hasattr(obj, '__dict__')



from scopeton import constants, compat


def get_annotations(instance):
    if compat.is_method(instance) or compat.is_function(instance):
        if hasattr(instance, constants.ANNOTATED):
            return getattr(instance, constants.ANNOTATED)
        return []

    return {fn: getattr(fn, constants.ANNOTATED) for fn in compat.get_methods(instance) if hasattr(fn, constants.ANNOTATED)}


def set_annotation(fn, name, value):
    if not hasattr(fn, constants.ANNOTATED):
        setattr(fn, constants.ANNOTATED, {})
    annotations = getattr(fn, constants.ANNOTATED)
    annotations[name] = value


def get_methods_with_annotation(instance, name):
    annotations = get_annotations(instance)
    return [k for k in annotations if name in annotations[k]]



from scopeton import glob, constants, annotation_tools


def Inject():
    def inject_decorator(fn):
        if fn.__name__ == '__init__':
            # constructor injector
            def inject_wrapper(*args, **kwags):
                return glob.lastScope._inject_method_args(fn, args[0])
        else:
            inject_wrapper = fn

        setattr(inject_wrapper, constants.TO_INJECT_FLAG, 1)
        return inject_wrapper

    return inject_decorator


def PostConstruct():
    def post_construct_decorator(fn):
        annotation_tools.set_annotation(fn, constants.POST_CONSTRUCT, 1)
        return fn

    return post_construct_decorator

def PreDestroy():
    def pre_destroy_decorator(fn):
        annotation_tools.set_annotation(fn, constants.PRE_DESTROY, 1)
        return fn

    return pre_destroy_decorator



import inspect
import sys


class _Python2NativeArgs:
    args = ['self']


def is_python3():
    return sys.version_info.major >= 3


def get_methods(instance):
    return [k[1] for k in inspect.getmembers(instance, predicate=inspect.ismethod)]


def get_method_instance(method):
    if not hasattr(method, "__self__"):
        return None
    return method.__self__


def is_class(obj):
    if not is_python3():
        import new
        return type(obj) is new.classobj or type(obj) is type
    else:
        import inspect
        return inspect.isclass(obj)


def object_is_instance(obj):
    if not is_python3():
        import new
        return type(obj) is new.instance or (hasattr(obj, "__class__") and hasattr(obj.__class__, "__name__"))
    else:
        import inspect
        return inspect.isclass(obj)


def get_method_signature(method):
    if is_python3():
        return inspect.getfullargspec(method)
    else:
        try:
            return inspect.getargspec(method)
        except TypeError:
            return _Python2NativeArgs()


def is_method(fn):
    return inspect.ismethod(fn)
def is_function(fn):
    return inspect.isfunction(fn)


import logging
import typing
from threading import RLock

from scopeton import compat, glob, constants, type_utils, scope_tools, annotation_tools
from scopeton.objects import Bean
from scopeton.qualifier_tree import QualifierTree
from scopeton.scope_tools import get_bean_qualifier, ScopetonException

T = typing.TypeVar("T")


class Scope(object):
    '''this is servicelocator pattern implementation'''

    def __init__(self, lock=False, parent=None):
        self._singletons = QualifierTree()
        self._beans = QualifierTree()
        self.lock = lock or RLock()  # type: RLock
        self.parent = parent  # type:  Scope
        self.servicesStarted = False
        self.children = []  # type: typing.List[Scope]
        self.registerInstance(self.__class__, self)
        if parent:
            parent.children.append(self)

    def getInstance(self, name: typing.Type[T]) -> T:
        if type_utils.is_collection(name):
            args = type_utils.get_type_args(name)
            return self.get_instances(typing.cast(typing.Type[T], args[0]))
        with self.lock:
            return self._get_instance(get_bean_qualifier(name))

    def get_instances(self, qualifier: typing.Type[T]) -> typing.List[T]:
        with self.lock:
            beans = self._beans.find_by_qualifier_name(get_bean_qualifier(qualifier))
            beans = map(lambda x: self.getInstance(x), beans)
            return list(beans)

    def _get_instance(self, qualifier):

        suitable_qualifier = self._beans.find_suitable_qualifier(qualifier)

        if self._singletons.contains(suitable_qualifier):
            return self._singletons.find_one_by_qualifier_name(suitable_qualifier)

        bean = self._beans.find_one_by_qualifier_name(suitable_qualifier)
        glob.lastScope = self

        if hasattr(bean.cls.__init__, constants.TO_INJECT_FLAG):
            # constructor injection
            instance = bean.cls()
        elif len(compat.get_method_signature(bean.cls.__init__).args) == 2:
            instance = bean.cls(self)
        elif len(compat.get_method_signature(bean.cls.__init__).args) > 2:
            raise ScopetonException(
                "Invalid number of parameters for bean constructor, maybe @Inject() decorator forgotten: {}".format(
                    compat.get_method_signature(bean.cls.__init__).args))
        else:
            instance = bean.cls()

        if bean.singleton:
            self.registerInstance(suitable_qualifier, instance)

        self._inject_injectables(instance)

        return instance

    def registerInstance(self, name, instance):
        scope_tools.set_instance_attr(instance, constants.SCOPE_PARAMETER, self)
        qualifier = get_bean_qualifier(name)
        suitableQualifier = self._beans.find_suitable_qualifier(qualifier)
        logging.debug("Suitable qualifier for {} is: {}".format(qualifier, suitableQualifier))
        self._singletons.register(suitableQualifier, instance)
        self._inject_injectables(instance)

    def registerBean(self, *args):
        with self.lock:
            for bean in args:
                if not isinstance(bean, Bean):
                    bean = Bean(bean)
                self._registerBean(bean)

    def remove(self):
        logging.debug("Removing scope: {}".format(self))
        self.stopServices()
        for k in self.children[:]:
            k.remove()
        if self.parent:
            self.parent.children.remove(self)

    def _registerBean(self, bean):
        """
        :type bean: Bean
        """
        for name in bean.qualifier_tree:
            logging.debug("Registering: {} as {}".format(name, bean))
            self._beans.register(name, bean)

    def runServices(self):
        if not self.servicesStarted:
            self.servicesStarted = True
            for bean in self._beans.get_all_objects():
                if bean.service:
                    instance = self.getInstance(bean)
                    methods = annotation_tools.get_methods_with_annotation(instance, constants.POST_CONSTRUCT)
                    for k in methods:
                        self._inject_method_args(k)

            for childScope in self.children:
                childScope.runServices()

    def stopServices(self):
        if self.servicesStarted:
            self.servicesStarted = False
            for bean in self._beans.get_all_objects():
                if bean.service:
                    methods = annotation_tools.get_methods_with_annotation(self.getInstance(bean), constants.PRE_DESTROY)
                    for k in methods:
                        self._inject_method_args(k)
            for childScope in self.children:
                childScope.stopServices()

    def _inject_injectables(self, instance):
        if hasattr(instance, constants.INJECTED):
            return
        for fn in scope_tools.get_injectables(instance):
            self._inject_method_args(fn)
        scope_tools.set_instance_attr(instance, constants.INJECTED, 1)

    def _inject_method_args(self, fn, add_self=None):
        signature = compat.get_method_signature(fn)
        nargs = []
        if add_self is not None:
            nargs.append(add_self)
        for arg_name in signature.args:
            if arg_name == "self":
                continue
            if arg_name not in signature.annotations:
                raise ScopetonException("Not annotated inject argument: {}".format(arg_name))
            nargs.append(self.getInstance(signature.annotations[arg_name]))
        return fn(*nargs)



import unittest

from scopeton import annotation_tools


class Dependency(object):
    def test(self):
        pass

    def testDecorated(self, param):
        pass


class AnnotationToolsTest(unittest.TestCase):

    def test_set_annotations(self):
        annotation_tools.set_annotation(Dependency.test, "test_annotation", 1)
        annotated_methods = annotation_tools.get_methods_with_annotation(Dependency(), "test_annotation")
        self.assertEqual(annotated_methods[0].__func__, Dependency.test)

    def test_get_annotations(self):
        annotation_tools.set_annotation(Dependency.test, "test_annotation", 1)
        annotated_methods = annotation_tools.get_annotations(Dependency.test)
        self.assertEqual(annotated_methods, {"test_annotation": 1})
        annotated_methods = annotation_tools.get_annotations(Dependency().test)
        self.assertEqual(annotated_methods, {"test_annotation": 1})



TO_INJECT_FLAG= "__inject__"
INJECTED = "__injected__"
ANNOTATED = "__anno__"
ANNOTATIONS = "__annotations__"
POST_CONSTRUCT = "__post_construct__"
PRE_DESTROY = "__pre_destroy__"
SCOPE_PARAMETER = "__scopeton_scope__"

